import yaml

from . import Parameter

class Parameters:
    """Collection of parameters.

    New instances may optionally be passed a ``path`` to a YAML file.
    See :meth:`.load_file`.

    .. code-block:: python

      parameters = Parameters()
      parameters.add_parameter('warp speed', 9.0)
      parameters.get_value('warp speed') #=> 9.0

      parameters.set_value('warp speed', 10.0)
      parameters.get_value('warp speed') #=> 10.0
    """

    def __init__(self, path=None):
        if path: self.load_file(path)

    @property
    def name(self):
        """The name of this parameter set.

        :return: Parameter set name.
        :rtype: str
        :default: None
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def property_keys(self):
        """Keys to load as class properties when loading from file.

        :return: Key names.
        :rtype: dict
        :default: ``['name']``
        """
        if not hasattr(self, '_property_keys'):
            self._property_keys = ['name']
        return self._property_keys

    @property_keys.setter
    def property_keys(self, value):
        self._property_keys = value

    @property
    def parameters(self):
        """The list of :class:`.Parameter` objects.

        :return: Parameter list.
        :rtype: dict
        :default: ``[]``
        """
        if not hasattr(self, '_parameters'): self._parameters = []
        return self._parameters

    @parameters.setter
    def parameters(self, value):
        self._parameters = value

    def get(self, name):
        """Get a parameter object by name.

        :param name: Name of the parameter object.
        :type name: str

        :return: The parameter.
        :rtype: Parameter
        """
        parameter = next((p for p in self.parameters if p.name == name), None)
        if parameter is None:
            raise LookupError("Cannot find parameter '" + name + "'.")
        return parameter

    def add_parameter(self, name, value, meta=None):
        """Add a parameter to the parameter list.

        :param name: New parameter's name.
        :type name: str

        :param value: New parameter's value.
        :type value: float

        :param meta: New parameter's meta property.
        :type meta: dict
        """
        parameter = Parameter(name, value)
        if meta: parameter.meta = meta
        self.parameters.append(parameter)

    def get_value(self, name):
        """Get the value of a parameter by name.

        :return: Parameter value.
        :rtype: Parameter
        """
        return self.get(name).value

    def set_value(self, name, value):
        """Set the value of a parameter by name.

        :param name: Parameter name.
        :type name: str

        :param value: Parameter value.
        :type value: float
        """
        self.get(name).value = value

    def get_meta(self, name):
        """Get the metadata of a parameter by name.

        :return: Parameter metadata.
        :rtype: dict
        """
        return self.get(name).meta

    def load_file(self, path):
        """Load a YAML file with parameter data and other metadata.

        :param path: Path to YAML file.
        :type path: str

        The data in the YAML file is used to set the properties for the instance.
        The YAML file must contain a ``parameters`` key.
        It may optionally include any keys defined in :attr:`.property_keys`.

        .. code-block:: yaml

          # data.yml
          ---
          name: Shield frequencies
          parameters:
            - name: a
              value: 24.50
            - name: β
              value: 42.10
              meta:
                phase_inverted: true


        .. code-block:: python

          parameters = Parameters('data.yml')
          parameters.name #=> 'Shield frequencies'
          parameters.get_value('a') #=> 24.50
          parameters.get_meta('β')['phase_inverted'] #=> true
        """
        data = yaml.load(open(path, 'r'))
        for key in self.property_keys:
            if key in data: setattr(self, key, data[key])
        self.parameters = self.parameter_list(data['parameters'])

    @staticmethod
    def parameter_list(data):
        """Create a list of parameter objects from a dict.

        :param data: Dictionary to convert to parameter list.
        :type data: dict

        :return: Parameter list.
        :rtype: dict
        """
        items = []
        for item in data:
            param = Parameter(item['name'], item['value'])
            if 'meta' in item: param.meta = item['meta']
            items.append(param)
        return items



from dichalcogenides.dichalcogenide import UpperValenceBand

from . import Superconductor

class Induced(Superconductor):
    """Induced superconductor.

    Inherits from :class:`.Superconductor`.
    """

    def __init__(self, model):
        super().__init__(model)
        self._energy = UpperValenceBand(self.model)

    @property
    def Δc(self):
        """Conduction band coupling constant.

        :return: :math:`Δ_c`
        :rtype: float
        """
        return self.model.system_parameter('Δc')

    @property
    def Δv(self):
        """Valance band coupling constant.

        :return: :math:`Δ_v`
        :rtype: float
        """
        return self.model.system_parameter('Δv')

    @property
    def Δk(self):
        """The superconducting parameter :math:`Δ_\\mathbf{k}`
        as a function of the energy :math:`E`.

        :return: :math:`Δ(E)`
        :rtype: func
        """
        Δc, Δv = self.Δc, self.Δv
        cos = self._energy.trig('cos θ')
        return lambda e: (1 / 2) * ((Δc + Δv) + (Δc - Δv) * cos(e))



import scipy.constants

from . import UpperValenceBand

class Optical():
    """Dichalcogenide properties related to optical transitions.

    Instantiate with a :class:`.Dichalcogenide` instance, e.g.,

    .. code-block:: python

       optical = Optical(Dichalcogenide('mos2', 'default'))
    """

    def __init__(self, dichalcogenide):
        self._dichalcogenide = dichalcogenide
        self._energy = UpperValenceBand(self._dichalcogenide)

    def p_circular(self, e, τ, ϵ):
        """Matrix element for optical transitions (circular polarization):
        :math:`\\left|P^{+-}_α \\left(E, \\mathbf{ϵ_±} \\right)\\right|^2`.

        Transitions are from the upper valance band to the lower conduction band.
        Given in units of :math:`\\text{GeV}^2`.

        :param e: Energy :math:`E`.
        :type e: float

        :param τ: Valley index :math:`τ = ±1`.
        :type τ: int

        :param ϵ: Polarization index :math:`ϵ = ±1`.
        :type ϵ: int
        """
        constant = scipy.constants.value

        at = self._dichalcogenide.at
        m0 = constant('electron mass energy equivalent in MeV')
        ℏ = constant('Planck constant over 2 pi in eV s')
        c = constant('speed of light in vacuum')
        Å = scipy.constants.angstrom

        sin = self._energy.trig('sin^2 θ/2')
        cos = self._energy.trig('cos^2 θ/2')
        x = 2 * (10**3 * at * Å * m0 / (c * ℏ))**2

        if ϵ * τ == 1:
            return x * sin(e)**2
        elif ϵ * τ == -1:
            return x * cos(e)**2
        else:
            raise RuntimeError(
                'Valley and polarization indexes must have values 1 or -1')



import numpy

class Energy():
    """Dichalcogenide properties related to the energy.

    Instantiate with a :class:`.Dichalcogenide` instance, e.g.,

    .. code-block:: python

       energy = Energy(Dichalcogenide('mos2', 'default'))
       energy.e(0, 1, 1, 1)
    """

    def __init__(self, dichalcogenide):
        self._dichalcogenide = dichalcogenide

    def e(self, k, n, τ, σ):
        """Band energy.

        :param k: Momentum :math:`k`.
        :type k: float

        :param n: Band index :math:`n = ±1`.
        :type n: int

        :param τ: Valley index :math:`τ = ±1`.
        :type τ: int

        :param σ: Spin index :math:`σ = ±1`.
        :type σ: int

        :return: :math:`E^n_{τ σ} (k)`
        :rtype: float
        """
        d = self._dichalcogenide
        at, Δ, λ = d.at, d.Δ, d.λ
        sqrt = numpy.sqrt
        α = τ * σ
        return 0.5 * (λ * α + n * sqrt((2 * at * k)**2 + (Δ - λ * α)**2))

    # TODO: Raise error when e is out of range given n.
    # pylint: disable=unused-argument
    def k(self, e, n, τ, σ):
        """Inverse energy-momentum relation.

        :param e: Energy :math:`E`.
        :type e: float

        :param n: Band index :math:`n = ±1`.
        :type n: int

        :param τ: Valley index :math:`τ = ±1`.
        :type τ: int

        :param σ: Spin index :math:`σ = ±1`.
        :type σ: int

        :return: :math:`k(E)`
        :rtype: float
        """
        d = self._dichalcogenide
        at, Δ, λ = d.at, d.Δ, d.λ
        sqrt = numpy.sqrt

        x = 2 * e * Δ**-1
        y = (λ * Δ**-1) * (1 - x)

        return (at**-1) * (Δ / 2) * sqrt(x**2 + 2 * τ * σ * y - 1)

class UpperValenceBand():
    """Dichalcogenide properties related to the upper valence band energy.

    For this class, the energy is assumed to be in the upper valence band:
    :math:`n = -1` and :math:`τ = σ`.

    Instantiate with a :class:`.Dichalcogenide` instance, e.g.,

    .. code-block:: python

       uvb = UpperValenceBand(Dichalcogenide('mos2', 'default'))
       uvb.ρ(0)
    """

    def __init__(self, dichalcogenide):
        self._dichalcogenide = dichalcogenide

    @property
    def Δμ(self):
        """Chemical potential offset.

        The distance of the chemical potential
        from the top of the upper valence band.

        :return: :math:`Δμ`
        :rtype: float
        :default: :math:`-λ`
        """
        if not hasattr(self, '_Δμ'): self.Δμ = -self._dichalcogenide.λ
        return self._Δμ

    @Δμ.setter
    def Δμ(self, value):
        self._Δμ = value

    @property
    def μ(self):
        """Chemical potential.

        Determined by the chemical potential offset.

        :return: :math:`μ = λ - Δ / 2 + Δμ`
        :rtype: float
        """
        d = self._dichalcogenide
        Δ, λ, Δμ = d.Δ, d.λ, self.Δμ
        return λ - Δ / 2 + Δμ

    @property
    def ξ_bounds(self):
        """Allowed range for the variable :math:`ξ = E - μ`.

        :return: :math:`(\\left|Δμ\\right| - 2 λ, \\left|Δμ\\right|)`
        :rtype: tuple
        """
        λ, Δμ = self._dichalcogenide.λ, self.Δμ
        return (abs(Δμ) - 2 * λ, abs(Δμ))

    def ρ(self, e):
        """Density of states.

        :param e: Energy :math:`E`.
        :type e: float

        :return: :math:`ρ(E) = \\left|k'(E) k(E)\\right|`
        :rtype: float
        """
        d = self._dichalcogenide
        at, λ = d.at, d.λ
        return abs(2 * e - λ) * (2 * at**2)**(-1)

    def trig(self, name):
        """Trigonometric functions of the energy.

        * ``sin θ``       - :math:`\\sin θ(E)`
        * ``cos θ``       - :math:`\\cos θ(E)`
        * ``sin^2 θ/2``   - :math:`\\sin^2 \\frac{θ(E)}{2}`
        * ``cos^2 θ/2``   - :math:`\\cos^2 \\frac{θ(E)}{2}`

        :param name: One of the above function names.
        :type name: str

        :return: Corresponding function of the energy :math:`E`.
        :rtype: function
        """
        fn = {}
        λ, Δ = self._dichalcogenide.λ, self._dichalcogenide.Δ
        sqrt = numpy.sqrt
        x = lambda e: (2 * e - λ)**(-1)

        fn['sin θ'] = lambda e: sqrt((2 * e - Δ) * (2 * e + Δ - 2 * λ)) * x(e)
        fn['cos θ'] = lambda e: (Δ - λ) * x(e)
        fn['sin^2 θ/2'] = lambda e: (e - Δ / 2) * x(e)
        fn['cos^2 θ/2'] = lambda e: (e + Δ / 2 - λ) * x(e)

        return fn[name]



import os

from . import Material
from . import System

class Model:
    """Combines material properties with system parameters.

    If, on instantiation, ``model`` or ``system`` is a string,
    the associated object will be created from the YAML file, e.g.,

    .. code-block:: python

       Model('mos2', 'default', 'data')

    will load the YAML files at ``data/mos2.yml`` and ``data/default.yml``.
    See :meth:`parameters.Parameters.load_file`.
    """

    def __init__(self, material=Material(), system=System(), root=''):
        if isinstance(material, str):
            self.material = Material(
                os.path.join(root, 'materials', material) + '.yml')
        else: self.material = material

        if isinstance(system, str):
            self.system = System(
                os.path.join(root, 'systems', system) + '.yml')
        else: self.system = system

    @property
    def material(self):
        """The associated :class:`.Material` object.

        :return: The material.
        :rtype: Material
        """
        return self._material

    @material.setter
    def material(self, value):
        self._material = value

    @property
    def system(self):
        """The associated :class:`.System` object.

        :return: The system.
        :rtype: System
        """
        return self._system

    @system.setter
    def system(self, value):
        self._system = value

    def material_parameter(self, name):
        """Get the value of a material parameter by name.

        :param name: Material name.
        :type name: str

        :return: The value.
        :rtype: float
        """
        return self.material.get_value(name)

    def system_parameter(self, name):
        """Get the value of a system parameter by name.

        :param name: System name.
        :type name: str

        :return: The value.
        :rtype: float
        """
        return self.system.get_value(name)



class Parameter:
    """Defines a parameter with a name and a value.

    Must provide a name and value.
    Parameters have a generic metadata property stored as a dictionary.

    .. code-block:: python

      parameter = Parameter('warp speed', 9.0, {'max': 10.0})
      parameter.name #=> 'warp speed'
      parameter.value #=> 9.0
      parameter.meta['max'] #=> 10.0
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

    @property
    def name(self):
        """The name of this parameter.

        :return: Parameter name.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def value(self):
        """The value of this parameter.

        :return: Parameter name.
        :rtype: float
        """
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def meta(self):
        """The metadata of this parameter.

        :return: Parameter metadata.
        :rtype: dict
        :default: ``{}``
        """
        if not hasattr(self, '_meta'): self._meta = {}
        return self._meta

    @meta.setter
    def meta(self, value):
        self._meta = value



class Topology():
    """Dichalcogenide topological properties.

    Instantiate with a :class:`.Dichalcogenide` instance, e.g.,

    .. code-block:: python

       topology = Topology(Dichalcogenide('mos2', 'default'))
       topology.Ω(0, 1, 1, 1)
    """

    def __init__(self, dichalcogenide):
        self._dichalcogenide = dichalcogenide

    def Ω(self, k, n, τ, σ):
        """Berry curvature.
        :param k: Momentum :math:`k`.
        :type k: float

        :param n: Band index :math:`n = ±1`.
        :type n: int

        :param τ: Valley index :math:`τ = ±1`.
        :type τ: int

        :param σ: Spin index :math:`σ = ±1`.
        :type σ: int

        :return: :math:`Ω^n_{τ σ} (k)`
        :rtype: float
        """
        d = self._dichalcogenide
        at, Δ, λ = d.at, d.Δ, d.λ
        x = Δ - λ * τ * σ
        y = (2 * at * k)**2 + x**2
        return -2 * (n * τ * at**2) * x * y**(-3/2)



import numpy

from dichalcogenides.dichalcogenide import UpperValenceBand

class Superconductor():
    """Superconductor.

    Uses upper valence band energies by default.
    """

    def __init__(self, model, energy=None):
        self.model = model
        if not energy: self._energy = UpperValenceBand(self.model)

    @property
    def model(self):
        """The model for this superconductor.

        :return: The model.
        :rtype: Model
        """
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def Δk(self):
        """The superconducting parameter :math:`Δ_\\mathbf{k}`
        as a function of the energy :math:`E`.

        :return: :math:`Δ(E)`
        :rtype: func
        """
        # pylint: disable=no-self-use
        raise NotImplementedError

    @property
    def μ(self):
        return self._energy.μ

    @property
    def λk_bounds(self):
        # TODO: Improve bound estimate using chemical potential.
        # ξ = self._energy.ξ_bounds
        return lambda Δk: (-10 * Δk, 10 * Δk)

    @property
    def ξ(self):
        return lambda Δk, λk: numpy.sqrt(λk**2 - Δk**2)

    def trig(self, name):
        ξ = self.ξ

        fn = {}
        fn['sin 2β'] = lambda Δk, λk: - abs(Δk) / λk
        fn['cos 2β'] = lambda Δk, λk: - ξ(Δk, λk) / λk
        fn['sin^2 β'] = lambda Δk, λk: (1 / 2) * (1 - ξ(Δk, λk) / λk)
        fn['cos^2 β'] = lambda Δk, λk: (1 / 2) * (1 + ξ(Δk, λk) / λk)

        return fn[name]

    def trig_e(self, name):
        μ = self.μ
        Δk = self.Δk
        λk = lambda e: numpy.sqrt((e - μ)**2 + Δk(e)**2)
        return lambda e: self.trig(name)(Δk(e), λk(e))



import os
import re
from setuptools import find_packages, setup

from dichalcogenides import __version__

with open('README.rst', 'r') as f:
    long_description = f.read()

if os.environ.get('READTHEDOCS') == 'True':
    mocked = ['numpy', 'scipy']
    mock_filter = lambda x: re.sub(r'>.+', '', x) not in mocked
else:
    mock_filter = lambda p: True

setup(
    name='dichalcogenides',
    version=__version__,
    author='Evan Sosenko',
    author_email='razorx@evansosenko.com',
    packages=find_packages(exclude=['docs']),
    url='https://github.com/razor-x/dichalcogenides',
    license='MIT',
    description='Python analysis code for dichalcogenide systems.',
    long_description=long_description,
    install_requires=list(filter(mock_filter, [
        'numpy>=1.11.0,<2.0.0',
        'scipy>=0.17.0,<1.0.0',
        'pyyaml>=3.11,<4.0'
    ]))
)



from dichalcogenides.model import Model

class Dichalcogenide(Model):
    """Model for a dichalcogenide.


    Inherits from :class:`model.Model`.
    """

    @property
    def at(self):
        """Product of the lattice constant and hopping energy.

        :return: :math:`at`
        :rtype: float
        """
        return self.material_parameter('at')

    @property
    def λ(self):
        """Half the valence band spin splitting.

        :return: :math:`λ`
        :rtype: float
        """
        return self.material_parameter('λ')

    @property
    def Δ(self):
        """Valance band conduction band energy splitting.

        :return: :math:`Δ`
        :rtype: float
        """
        return self.material_parameter('Δ')
